// Copyright 2016 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
// Modified by Jeremy Retailleau.

#ifndef PXR_VT_TYPE_HEADERS_H
#define PXR_VT_TYPE_HEADERS_H

#include <string>

#include "pxr/vt/array.h"

#include <pxr/tf/token.h>

#include <pxr/gf/dualQuatd.h>
#include <pxr/gf/dualQuatf.h>
#include <pxr/gf/dualQuath.h>
#include <pxr/gf/frustum.h>
#include <pxr/gf/half.h>
#include <pxr/gf/interval.h>
#include <pxr/gf/multiInterval.h>
#include <pxr/gf/matrix2f.h>
#include <pxr/gf/matrix3f.h>
#include <pxr/gf/matrix4f.h>
#include <pxr/gf/matrix2d.h>
#include <pxr/gf/matrix3d.h>
#include <pxr/gf/matrix4d.h>
#include <pxr/gf/quatd.h>
#include <pxr/gf/quatf.h>
#include <pxr/gf/quath.h>
#include <pxr/gf/quaternion.h>
#include <pxr/gf/range1d.h>
#include <pxr/gf/range1f.h>
#include <pxr/gf/range2d.h>
#include <pxr/gf/range2f.h>
#include <pxr/gf/range3d.h>
#include <pxr/gf/range3f.h>
#include <pxr/gf/rect2i.h>
#include <pxr/gf/vec2d.h>
#include <pxr/gf/vec2f.h>
#include <pxr/gf/vec2h.h>
#include <pxr/gf/vec2i.h>
#include <pxr/gf/vec3d.h>
#include <pxr/gf/vec3f.h>
#include <pxr/gf/vec3h.h>
#include <pxr/gf/vec3i.h>
#include <pxr/gf/vec4d.h>
#include <pxr/gf/vec4f.h>
#include <pxr/gf/vec4h.h>
#include <pxr/gf/vec4i.h>

#endif // PXR_VT_TYPE_HEADERS_H
