// Copyright 2025 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
// Modified by Jeremy Retailleau.

#include "pxr/vt/pxr.h"
#include <pxr/tf/debug.h>

VT_NAMESPACE_OPEN_SCOPE

TF_DEBUG_CODES(
    VT_ARRAY_EDIT_BOUNDS
);

VT_NAMESPACE_CLOSE_SCOPE
