// Copyright 2016 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
// Modified by Jeremy Retailleau.

#include <pxr/vt/pxr.h>
#include <pxr/vt/wrapArray.h>

VT_NAMESPACE_USING_DIRECTIVE

void wrapArray()
{
    // The actual wrapping of particular template instantiations is done in the
    // specific wrapArrayXXX.cpp files to avoid quadratic compiler behavior.
}
